library pointycastle.impl.stream_cipher.chacha208;

import "dart:typed_data";

import 'package:steel_crypt/PointyCastleN/export.dart';

import '../api.dart';
import '../src/impl/base_stream_cipher.dart';
import '../src/registry/registry.dart';
import '../src/ufixnum.dart';

/// Implementation of Daniel J. Bernstein's ChaCha20 stream cipher, Snuffle 2005.
class ChaCha8Engine extends BaseStreamCipher {
  static final FactoryConfig FACTORY_CONFIG =
  StaticFactoryConfig(StreamCipher, "ChaCha20/8", () => ChaCha8Engine());

  static const STATE_SIZE = 16;

  static final _sigma = Uint8List.fromList("expand 32-byte k".codeUnits);
  static final _tau = Uint8List.fromList("expand 16-byte k".codeUnits);

  Uint8List _workingKey;
  Uint8List _workingIV;

  final _state = List<int>(STATE_SIZE);
  final _buffer = List<int>(STATE_SIZE);

  final _keyStream = Uint8List(STATE_SIZE * 4);
  var _keyStreamOffset = 0;

  var _initialised = false;

  final String algorithmName = "ChaCha20";

  void reset() {
    if (_workingKey != null) {
      _setKey(_workingKey, _workingIV);
    }
  }

  void init(
      bool forEncryption, covariant ParametersWithIV<KeyParameter> params) {
    var uparams = params.parameters;
    var iv = params.iv;
    if (iv == null || iv.length != 8) {
      throw ArgumentError("ChaCha20 requires exactly 8 bytes of IV");
    }

    _workingIV = iv;
    _workingKey = uparams.key;

    _setKey(_workingKey, _workingIV);
  }

  int returnByte(int inp) {
    if (_keyStreamOffset == 0) {
      _generateKeyStream(_keyStream);

      if (++_state[12] == 0) {
        ++_state[13];
      }
    }

    var out = clip8(_keyStream[_keyStreamOffset] ^ inp);
    _keyStreamOffset = (_keyStreamOffset + 1) & 63;

    return out;
  }

  void processBytes(
      Uint8List inp, int inpOff, int len, Uint8List out, int outOff) {
    if (!_initialised) {
      throw StateError("ChaCha20 not initialized: please call init() first");
    }

    if ((inpOff + len) > inp.length) {
      throw ArgumentError(
          "Input buffer too short or requested length too long");
    }

    if ((outOff + len) > out.length) {
      throw ArgumentError(
          "Output buffer too short or requested length too long");
    }

    for (var i = 0; i < len; i++) {
      if (_keyStreamOffset == 0) {
        _generateKeyStream(_keyStream);

        if (++_state[12] == 0) {
          ++_state[13];
        }
      }

      out[i + outOff] = clip8(_keyStream[_keyStreamOffset] ^ inp[i + inpOff]);
      _keyStreamOffset = (_keyStreamOffset + 1) & 63;
    }
  }

  void _setKey(Uint8List keyBytes, Uint8List ivBytes) {
    _workingKey = keyBytes;
    _workingIV = ivBytes;

    _keyStreamOffset = 0;
    int offset = 0;
    Uint8List constants;

    // Key
    _state[4] = unpack32(_workingKey, 0, Endian.little);
    _state[5] = unpack32(_workingKey, 4, Endian.little);
    _state[6] = unpack32(_workingKey, 8, Endian.little);
    _state[7] = unpack32(_workingKey, 12, Endian.little);

    if (_workingKey.length == 32) {
      constants = _sigma;
      offset = 16;
    } else {
      constants = _tau;
    }

    _state[8] = unpack32(_workingKey, offset, Endian.little);
    _state[9] = unpack32(_workingKey, offset + 4, Endian.little);
    _state[10] = unpack32(_workingKey, offset + 8, Endian.little);
    _state[11] = unpack32(_workingKey, offset + 12, Endian.little);
    _state[0] = unpack32(constants, 0, Endian.little);
    _state[1] = unpack32(constants, 4, Endian.little);
    _state[2] = unpack32(constants, 8, Endian.little);
    _state[3] = unpack32(constants, 12, Endian.little);

    // IV
    _state[14] = unpack32(_workingIV, 0, Endian.little);
    _state[15] = unpack32(_workingIV, 4, Endian.little);
    _state[12] = _state[13] = 0;

    _initialised = true;
  }

  void _generateKeyStream(Uint8List output) {
    _core(8, _state, _buffer);
    var outOff = 0;
    for (var x in _buffer) {
      pack32(x, output, outOff, Endian.little);
      outOff += 4;
    }
  }

  /// The ChaCha20 core function
  void _core(int rounds, List<int> input, List<int> x) {
    int x00 = input[0];
    int x01 = input[1];
    int x02 = input[2];
    int x03 = input[3];
    int x04 = input[4];
    int x05 = input[5];
    int x06 = input[6];
    int x07 = input[7];
    int x08 = input[8];
    int x09 = input[9];
    int x10 = input[10];
    int x11 = input[11];
    int x12 = input[12];
    int x13 = input[13];
    int x14 = input[14];
    int x15 = input[15];

    for (int i = rounds; i > 0; i -= 2)
    {
      x00 += x04; x12 = crotl32(x12 ^ x00, 16);
      x08 += x12; x04 = crotl32(x04 ^ x08, 12);
      x00 += x04; x12 = crotl32(x12 ^ x00, 8);
      x08 += x12; x04 = crotl32(x04 ^ x08, 7);
      x01 += x05; x13 = crotl32(x13 ^ x01, 16);
      x09 += x13; x05 = crotl32(x05 ^ x09, 12);
      x01 += x05; x13 = crotl32(x13 ^ x01, 8);
      x09 += x13; x05 = crotl32(x05 ^ x09, 7);
      x02 += x06; x14 = crotl32(x14 ^ x02, 16);
      x10 += x14; x06 = crotl32(x06 ^ x10, 12);
      x02 += x06; x14 = crotl32(x14 ^ x02, 8);
      x10 += x14; x06 = crotl32(x06 ^ x10, 7);
      x03 += x07; x15 = crotl32(x15 ^ x03, 16);
      x11 += x15; x07 = crotl32(x07 ^ x11, 12);
      x03 += x07; x15 = crotl32(x15 ^ x03, 8);
      x11 += x15; x07 = crotl32(x07 ^ x11, 7);
      x00 += x05; x15 = crotl32(x15 ^ x00, 16);
      x10 += x15; x05 = crotl32(x05 ^ x10, 12);
      x00 += x05; x15 = crotl32(x15 ^ x00, 8);
      x10 += x15; x05 = crotl32(x05 ^ x10, 7);
      x01 += x06; x12 = crotl32(x12 ^ x01, 16);
      x11 += x12; x06 = crotl32(x06 ^ x11, 12);
      x01 += x06; x12 = crotl32(x12 ^ x01, 8);
      x11 += x12; x06 = crotl32(x06 ^ x11, 7);
      x02 += x07; x13 = crotl32(x13 ^ x02, 16);
      x08 += x13; x07 = crotl32(x07 ^ x08, 12);
      x02 += x07; x13 = crotl32(x13 ^ x02, 8);
      x08 += x13; x07 = crotl32(x07 ^ x08, 7);
      x03 += x04; x14 = crotl32(x14 ^ x03, 16);
      x09 += x14; x04 = crotl32(x04 ^ x09, 12);
      x03 += x04; x14 = crotl32(x14 ^ x03, 8);
      x09 += x14; x04 = crotl32(x04 ^ x09, 7);

    }
    List<int> xup = [x00, x01, x02, x03, x04, x05, x06, x07, x08, x09, x10, x11, x12, x13, x14, x15];
    for (int i = 0; i < STATE_SIZE; ++i) {
      x[i] = csum32(xup[i], input[i]);
    }
  }

  external dynamic noSuchMethod(Invocation invocation);
}